"""Extract visualizations of DINO features with PCA."""

import logging
from pathlib import Path

import h5py
import numpy as np


def _process_file(
    file_name: str,
    label_dict: dict[str, Path],
    result_dir: Path,
    threshold: float = 0.5,
) -> None:
    import cv2
    import seaborn as sns

    colors = sns.color_palette("deep")[:4]
    hue_palette = {
        "mito": colors[0],
        "cristae": colors[1],
        "microtubule": colors[2],
        "granule": colors[3],
    }

    label_data = {}
    sample = "unknown"
    for label, f_path in label_dict.items():
        sample = f_path.parent.name
        with h5py.File(f_path, "r") as fh:
            if "data" not in label_data:
                label_data["data"] = fh["data"][()].astype(np.float32)  # type: ignore # [0, 1], [D, H, W]
            label_data[label] = fh["predictions"][label][()].astype(np.float32)  # type: ignore # [0, 1], [D, H, W]
    # Combine all labels into a single segmentation map
    combined_rgb_seg = np.zeros(
        (*label_data["data"].shape, 3), dtype=np.float32
    )  # [D, H, W, 3]
    for label, seg in label_data.items():
        if label in hue_palette:
            color = np.array(hue_palette.get(label, (1.0, 1.0, 1.0))).reshape(
                (1, 1, 1, 3)
            )  # default to white if label not in palette
            rgb_seg = np.stack([seg, seg, seg], axis=-1)  # to RGB
            rgb_seg = rgb_seg * color  # colorize
            combined_rgb_seg += rgb_seg
        else:
            logging.warning("Couldn't find color for label %s", label)
    combined_rgb_seg = np.clip(combined_rgb_seg, 0, 1)
    label_data["data"] = np.clip(label_data["data"], 0, 1)
    final_data = np.stack(
        [label_data["data"], label_data["data"], label_data["data"]], axis=-1
    )  # to RGB
    # Add in data and normalize
    final_rgb = np.where(
        combined_rgb_seg > threshold, combined_rgb_seg, final_data
    )
    # Combine and convert to cv2 format
    final_combined = np.concatenate(
        [final_data, final_rgb], axis=2
    )  # [D, H, 2*W, 3]
    final_combined = (final_combined * 255).astype(np.uint8)
    frames = []
    for frame in final_combined:
        frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)  # to BGR for cv2
        frames.append(frame)
    # Save as Video
    result_path = result_dir / sample / (file_name + ".mp4")
    result_path.parent.mkdir(parents=True, exist_ok=True)
    video_path = str(result_path.with_suffix(".mp4"))
    fourcc = cv2.VideoWriter.fourcc(*"mp4v")
    fps = 30
    width, height = final_combined.shape[2], final_combined.shape[1]
    video_writer = cv2.VideoWriter(
        str(video_path), fourcc, fps, (width, height)
    )
    for img_frame in frames:
        video_writer.write(img_frame)
    video_writer.release()
    logging.info("Saved video to %s", video_path)


def _process_files(
    file_dict: dict[str, dict[str, Path]], result_dir: Path
) -> None:
    result_dir.mkdir(parents=True, exist_ok=True)
    for f_name, labels in file_dict.items():
        logging.info("Processing file %s", f_name)
        _process_file(f_name, labels, result_dir)


def process_experiment(
    exp_dir: Path,
    result_dir: Path,
    exp_template: str,
    labels: list[str] | None,
) -> None:
    """Process segmentation results from multiple labels and save combined visualizations as a video.

    Args:
        exp_dir (Path): Directory containing experiment results
        result_dir (Path): Directory to save results
        exp_template (str): String to identify relevant experiment directories (i.e., start with exp_template), where the label is assumed to be the suffix after the last underscore.
        labels (list[str] | None): List of labels to process. If None, all labels found based on the `exp_template` will be processed.
    """

    import seaborn as sns

    colors = sns.color_palette("deep")[:4]
    hue_palette = {
        "mito": colors[0],
        "cristae": colors[1],
        "microtubule": colors[2],
        "granule": colors[3],
    }

    result_dir.mkdir(parents=True, exist_ok=True)

    if labels is None:
        exp_names = [
            d.name
            for d in exp_dir.iterdir()
            if d.is_dir()
            and d.name.startswith(exp_template)
            and d.name.split("_")[-1] in hue_palette
        ]
        labels = [name.split("_")[-1] for name in exp_names]
    else:
        exp_names = [
            d.name
            for d in exp_dir.iterdir()
            if d.is_dir()
            and d.name.startswith(exp_template)
            and d.name.split("_")[-1] in labels
        ]

    files = {}
    for label, exp_name in zip(labels, exp_names, strict=True):
        file_dir = exp_dir / exp_name
        tomo_files = list(file_dir.glob("**/*.hdf"))
        logging.info(
            "Found %d .hdf files for label %s in experiment directory %s",
            len(tomo_files),
            label,
            exp_name,
        )
        files[label] = tomo_files
    all_files = {}
    for label, fs in files.items():
        for f in fs:
            if f.stem not in all_files:
                all_files[f.stem] = {label: f.resolve()}
            else:
                all_files[f.stem][label] = f.resolve()
    # Now, all_files maps each file name to a dictionary of labels and corresponding file paths for that label
    _process_files(
        all_files,
        result_dir / f"{exp_template}_{'_'.join(labels)}_segmentations",
    )



"""Implementation of the fractional leave one out data module."""

import pandas as pd

from cryovit.datamodules.base_datamodule import BaseDataModule


class FractionalSampleDataModule(BaseDataModule):
    """Data module for fractional leave-one-out CryoVIT experiments."""

    def __init__(
        self,
        sample: list[str],
        split_id: int | None,
        split_key: str | None,
        test_sample: list[str] | None = None,
        **kwargs,
    ) -> None:
        """Train on a fraction of tomograms and leave out one sample for evaluation.

        Args:
            sample (list[str]): The samples to train and test on.
            split_id (Optional[int]): The number of splits used for training. If None, defaults to all splits.
            split_key (str): The key used to select splits using split_id.
            test_sample (Optional[list[str]]): The sample to exclude from training and use for testing.
        """

        super().__init__(**kwargs)
        # Validity checks
        assert (
            test_sample is not None
        ), "Fractional sample `test_sample` cannot be None."
        assert (
            len(test_sample) == 1
        ), f"Fractional sample 'test_sample' should be a single string list. Got {test_sample} instead."

        self.sample = sample
        self.split_id = split_id
        self.split_key = split_key
        self.test_sample = test_sample

    def train_df(self) -> pd.DataFrame:
        """Train tomograms: include a subset of all splits, leaving out one sample.

        Returns:
            pd.DataFrame: A dataframe specifying the train tomograms.
        """

        assert self.record_df is not None
        if self.split_id is not None:
            training_splits = list(range(self.split_id))
        else:
            training_splits = list(range(self.record_df[self.split_key].max()))

        return self.record_df[
            (self.record_df[self.split_key].isin(training_splits))
            & (self.record_df["sample"].isin(self.sample))
            & ~self.record_df["sample"].isin(self.test_sample)
        ][["sample", "tomo_name"]]

    def val_df(self) -> pd.DataFrame:
        """Validation tomograms: validate on tomograms from the held out sample.

        Returns:
            pd.DataFrame: A dataframe specifying the validation tomograms.
        """

        assert self.record_df is not None

        return self.record_df[
            (self.record_df["sample"].isin(self.test_sample))
        ]

    def test_df(self) -> pd.DataFrame:
        """Test tomograms: test on tomograms from the held out sample.

        Returns:
            pd.DataFrame: A dataframe specifying the test tomograms.
        """

        assert self.record_df is not None
        if self.split_id is not None:
            keys = ["sample", "tomo_name", self.split_key]
        else:
            keys = ["sample", "tomo_name"]

        df = self.val_df()[keys]
        # replace split_id with split fraction if it exists
        if self.split_key in df.columns:
            df["split_id"] = self.split_id
        return df

    def predict_df(self) -> pd.DataFrame:
        """Predict tomograms: predict on the specified samples.

        Returns:
            pd.DataFrame: A dataframe specifying the predict tomograms.
        """

        assert self.record_df is not None
        return self.record_df[self.record_df["sample"].isin(self.sample)][
            ["sample", "tomo_name"]
        ]



from pathlib import Path
from typing import Annotated

from typer import Argument, Option

from .cli import cli


@cli.command(name="evaluate", no_args_is_help=True)
def evaluate(
    test_data: Annotated[
        str,
        Argument(
            help="Path to the folder or .txt file containing the test tomograms.",
        ),
    ],
    test_labels: Annotated[
        str,
        Argument(
            help="Path to the folder or .txt file containing the test labels.",
        ),
    ],
    labels: Annotated[
        list[str],
        Option(help="List of available label names in ascending-value order."),
    ],
    model: Annotated[
        str,
        Argument(
            help="Path to the .model file containing the pre-trained model.",
        ),
    ],
    result_folder: Annotated[
        str | None,
        Option(
            help="Path to the directory to save the evaluation results. Evaluation metrics will be saved to a .csv file in a folder named 'results' inside the result folder.",
            show_default="the current working directory",
        ),
    ] = None,
    visualize: Annotated[
        bool,
        Option(
            "--visualize",
            "-v",
            help="Save visualizations of model predictions?. This will slightly increase the runtime. Results will be saved in a folder named `predictions` inside the result folder.",
        ),
    ] = False,
):
    """Evaluate a pre-trained model on a test dataset.

    Example
    -------
    cryovit evaluate <path-to-test-data> <path-to-test-labels> --labels <first-label> ... --model <path-to-model-file> --result-folder <path-to-result-folder>
    """
    from cryovit._logging_config import setup_logging
    from cryovit.run.eval_model import run_evaluation
    from cryovit.utils import load_files_from_path, load_model

    setup_logging("INFO")

    ## Convert Arguments
    test_path = Path(test_data)
    label_path = Path(test_labels)
    model_path = Path(model)
    result_path = Path(result_folder) if result_folder else Path.cwd()

    ## Sanity Checking
    assert test_path.exists(), "Test data path does not exist."
    assert label_path.exists(), "Test labels path does not exist."
    assert (
        model_path.exists() and model_path.suffix == ".model"
    ), "Model path does not exist, or is not a .model file."
    _, _, _, label_key = load_model(model_path, load_model=False)
    assert (
        label_key in labels
    ), f"The label key {label_key} used to train the model is not in the provided labels."
    result_path.mkdir(parents=True, exist_ok=True)

    test_files = load_files_from_path(test_path)
    label_files = load_files_from_path(label_path)

    run_evaluation(
        test_files,
        label_files,
        labels,
        model_path,
        result_path,
        visualize=visualize,
    )



"""Metrics for evaluating CryoVIT segmentation models."""

import torch
from torch import Tensor
from torchmetrics import Metric


class DiceMetric(Metric):
    """Metric class for calculating the Dice score to evaluate segmentation models."""

    higher_is_better = True

    def __init__(self, threshold: float, **kwargs):
        """Initializes the DiceMetric instance with a threshold for binary classification.

        Args:
            threshold (float): The threshold to apply to the predictions during Dice score calculation.
        """

        super().__init__()
        self.name = "DiceMetric"
        self.thresh = threshold
        self.add_state(
            "dice_score", default=torch.tensor(0.0), dist_reduce_fx="sum"
        )
        self.add_state(
            "total", default=torch.tensor(0.0), dist_reduce_fx="sum"
        )

    def update(self, y_pred: Tensor, y_true: Tensor) -> None:
        """Updates the states for the Dice score computation based on predictions and actual values.

        Args:
            y_pred (Tensor): Predicted probabilities or logits from the model.
            y_true (Tensor): Ground truth labels.
        """

        y_pred = torch.where(y_pred < self.thresh, 0.0, 1.0)

        intersection = torch.sum(y_true * y_pred)
        denom = torch.sum(y_true) + torch.sum(y_pred)

        self.dice_score += 2 * intersection / (denom + 1e-3)
        self.total += 1

    def compute(self) -> Tensor:
        """Computes the final Dice score over all updates.

        Returns:
            Tensor: The average Dice score across all batches.
        """

        return self.dice_score / self.total if self.total > 0 else torch.tensor(0.0)  # type: ignore


class F1Metric(Metric):
    """Metric class for calculating the F1 score to evaluate segmentation models."""

    def __init__(self, **kwargs):
        """Initializes the F1Metric instance."""

        super().__init__()
        self.name = "F1Metric"
        self.add_state("f1", default=torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state(
            "total", default=torch.tensor(0.0), dist_reduce_fx="sum"
        )

    def update(self, y_pred: Tensor, y_true: Tensor) -> None:
        """Updates the states for the F1 score computation based on predictions and actual values.

        Args:
            y_pred (Tensor): Predicted probabilities or logits from the model.
            y_true (Tensor): Ground truth labels.
        """
        y_pred = (y_pred > 0.5).float()
        tp = torch.sum(y_true * y_pred)
        fp = torch.sum((1 - y_true) * y_pred)
        fn = torch.sum(y_true * (1 - y_pred))
        precision = tp / (tp + fp + 1e-6)
        recall = tp / (tp + fn + 1e-6)
        f1 = 2 * (precision * recall) / (precision + recall + 1e-6)
        self.f1 += f1
        self.total += 1

    def compute(self) -> Tensor:
        """Computes the final F1 score over all updates.

        Returns:
            Tensor: The average F1 score across all batches.
        """

        return self.f1 / self.total if self.total > 0 else torch.tensor(0.0)  # type: ignore



"""Implementation of the fractional  data module."""

import numpy as np
import pandas as pd
from sklearn.model_selection import KFold

from cryovit.datamodules.base_datamodule import BaseDataModule


class FractionalDataModule(BaseDataModule):
    """Data module for fractional CryoVIT experiments."""

    def __init__(
        self,
        sample: list[str],
        split_id: int | None,
        split_key: str | None,
        test_sample: int | None = None,
        **kwargs,
    ) -> None:
        """Train on a fraction of tomograms and leave out one sample for evaluation.

        Args:
            sample (list[str]): The samples to train and test on.
            split_id (Optional[int]): The number of splits used for training. If None, defaults to all splits.
            split_key (str): The key used to select splits using split_id.
            test_sample (Optional[int]): The split to exclude from training and use for testing.
        """

        super().__init__(**kwargs)
        # Validity checks
        assert (
            test_sample is not None
        ), "Fractional sample `test_sample` cannot be None."
        assert isinstance(
            test_sample, int
        ), f"Fractional sample 'test_sample' should be an integer. Got {test_sample} instead."

        # Add new splits for fractional training
        num_samples = self.record_df.shape[0]
        n_splits = 11  # Using 10-fold + 1 for LOO
        kf = KFold(n_splits=n_splits, shuffle=True, random_state=42)
        X = np.array([[0] for _ in range(num_samples)])
        splits = [-1 for _ in range(num_samples)]
        for f, (_, test) in enumerate(kf.split(X)):
            for idx in test:
                splits[idx] = f
        self.record_df[split_key] = splits

        self.sample = sample
        self.split_id = split_id
        self.split_key = split_key
        self.test_id = test_sample

    def train_df(self) -> pd.DataFrame:
        """Train tomograms: include a subset of all splits, leaving out one split.

        Returns:
            pd.DataFrame: A dataframe specifying the train tomograms.
        """

        assert self.record_df is not None
        all_splits = sorted(
            set(self.record_df[self.split_key].unique()) - {self.test_id}
        )
        assert (
            len(all_splits) == 10
        ), "There should be 10 splits for fractional training."
        if self.split_id is not None:
            training_splits = all_splits[: self.split_id]
        else:
            training_splits = all_splits

        return self.record_df[
            (self.record_df[self.split_key].isin(training_splits))
            & (self.record_df["sample"].isin(self.sample))
        ][["sample", "tomo_name"]]

    def val_df(self) -> pd.DataFrame:
        """Validation tomograms: validate on tomograms from the held out sample.

        Returns:
            pd.DataFrame: A dataframe specifying the validation tomograms.
        """

        assert self.record_df is not None
        return self.record_df[
            (self.record_df[self.split_key] == self.test_id)
            & (self.record_df["sample"].isin(self.sample))
        ]

    def test_df(self) -> pd.DataFrame:
        """Test tomograms: test on tomograms from the held out sample.

        Returns:
            pd.DataFrame: A dataframe specifying the test tomograms.
        """

        assert self.record_df is not None
        if self.split_id is not None:
            keys = ["sample", "tomo_name", self.split_key]
        else:
            keys = ["sample", "tomo_name"]

        df = self.val_df()[keys]
        # replace split_id with split fraction if it exists
        if self.split_key in df.columns:
            df["split_id"] = self.split_id
        return df

    def predict_df(self) -> pd.DataFrame:
        """Predict tomograms: predict on the specified samples.

        Returns:
            pd.DataFrame: A dataframe specifying the predict tomograms.
        """

        assert self.record_df is not None
        return self.record_df[self.record_df["sample"].isin(self.sample)][
            ["sample", "tomo_name"]
        ]



"""SAMv2 model for 2D/3D tomogram segmentation with a prompt predictor for automated segmentation.

Code is based on the original SAM2 training code from https://github.com/facebookresearch/sam2/blob/main/training/model/sam2.py.
"""

import logging
from pathlib import Path
from typing import Any, Literal

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from hydra.core.global_hydra import GlobalHydra
from hydra.utils import instantiate
from omegaconf import OmegaConf
from sam2.modeling.sam2_base import SAM2Base
from torch import Tensor
from torch.optim import Optimizer

from cryovit.config import BaseModel as BaseModelConfig
from cryovit.models.base_model import BaseModel
from cryovit.models.sam2_blocks import LoRAMaskDecoderFactory, PromptPredictor
from cryovit.types import BatchedTomogramData

# Clear SAM2 hydra initialization if it exists
if GlobalHydra.instance().is_initialized():
    GlobalHydra.instance().clear()

## Pre-trained Model Weights ##
sam2_model = (
    "facebook/sam2.1-hiera-tiny",
    {"config": "sam2.1_hiera_t.yaml", "weights": "sam2.1_hiera_tiny.pt"},
)  # the tiny variant of SAMv2.1
medical_sam2_model = (
    "wanglab/MedSAM2",
    {"config": "sam2.1_hiera_t.yaml", "weights": "MedSAM2_latest.pt"},
)  # fine-tuned on medical data SAMv2

MAX_SAM_DEPTH = 255  # Temporary maximum depth (number of slices) for SAMv2 (due to implementation error in CUDA - https://github.com/pytorch/pytorch/issues/142228)
# a large negative value as a placeholder score for missing objects
NO_OBJ_SCORE = -1024.0


class SAM2(BaseModel):
    """Lightning wrapper over the SAM2 model."""

    def __init__(
        self, sam_model: "SAM2Train", custom_kwargs, **kwargs
    ) -> None:
        """Initializes the SAM2 model with specific convolutional and synthesis blocks."""

        super().__init__(**kwargs)
        self.prompt_lr = custom_kwargs.get("prompt_lr", 3e-05)
        if "prompt_lr" in custom_kwargs:
            del custom_kwargs["prompt_lr"]
        self.model = sam_model(**custom_kwargs)
        self.prompt_predictor = PromptPredictor()

        self.freeze_parameters()
        self.log_masks = False  # whether to log predicted masks during training for debugging

    def freeze_parameters(self):
        """Freezes all model parameters except for the prompt predictor and mask decoder."""

        for p in self.model.image_encoder.parameters():
            p.requires_grad = False
        for p in self.model.sam_prompt_encoder.parameters():
            p.requires_grad = False
        for p in self.model.memory_encoder.parameters():
            p.requires_grad = False
        for p in self.model.memory_attention.parameters():
            p.requires_grad = False

    def configure_optimizers(self) -> Optimizer:
        """Configures the optimizer with separate learning rates for the prompt predictor and mask decoder."""

        prompt_params = {
            "params": self.prompt_predictor.parameters(),
            "lr": self.prompt_lr,
        }
        decoder_params = {
            "params": self.model.parameters(),
        }
        return torch.optim.AdamW(
            [prompt_params, decoder_params],
            lr=self.lr,
            weight_decay=self.weight_decay,
        )

    def _masked_predict(
        self,
        batch: BatchedTomogramData,  # type: ignore
        use_mito_mask: bool = False,
    ) -> dict[str, Tensor]:
        """Override trainer _masked_predict to handle masking for the prompt predictor."""

        out = self(batch)
        y_true = batch.labels  # (B, D, H, W)
        y_pred_full, mask_pred_full = out["preds"], out["prompts"]

        mask = (y_true > -1.0).detach()
        if use_mito_mask:
            assert (
                batch.aux_data is not None and "labels/mito" in batch.aux_data
            ), "Batch aux_data must contain 'labels/mito' key for mito masking."
            # assumes eval code with batch size of 1
            mito_mask = torch.from_numpy(batch.aux_data["labels/mito"][0]) > 0
            mito_mask = mito_mask.to(dtype=mask.dtype, device=mask.device)
            mask = mask & mito_mask  # Combine masks

        y_pred = torch.masked_select(y_pred_full, mask).view(-1, 1)
        y_true = torch.masked_select(y_true, mask).view(-1, 1)
        mask_pred = torch.masked_select(mask_pred_full, mask).view(-1, 1)

        return {
            "preds": y_pred,
            "masks": mask_pred,
            "labels": y_true,
            "preds_full": y_pred_full,
            "masks_full": mask_pred_full,
        }

    def _do_step(
        self,
        batch: BatchedTomogramData,  # type: ignore
        batch_idx: int,
        prefix: Literal["train", "val", "test"],
    ) -> Tensor:
        """Override trainer do_step to handle losses for the prompt predictor."""

        out_dict = self._masked_predict(batch)

        y_pred, y_true = out_dict["preds"], out_dict["labels"]

        losses = self.compute_losses(y_pred, y_true)
        mask_loss = self.compute_losses(out_dict["masks"], y_true)["dice_loss"]
        losses["mask_loss"] = mask_loss
        losses["total"] = losses["total"] + mask_loss

        for _, m_fn in self.metric_fns[prefix.upper()].items():  # type: ignore
            m_fn(y_pred, y_true)

        self.log_stats(losses, prefix, batch.num_tomos)
        if self.training and self.log_masks:
            import wandb

            # debug logging predicted masks
            raw_image = (
                batch.tomo_batch[0, batch.num_slices // 2].detach().cpu()[[0]]
            )
            pred_image = (
                out_dict["preds_full"][0, batch.num_slices // 2]
                .detach()
                .cpu()
                .unsqueeze(0)
            )
            mask_image = (
                out_dict["masks_full"][0, batch.num_slices // 2]
                .detach()
                .cpu()
                .unsqueeze(0)
            )
            prompt_image = (mask_image > 0.9).float()
            combined_image = torch.cat(
                [raw_image, pred_image, mask_image, prompt_image], dim=-1
            )  # concat along width
            wandb.log(
                {
                    "raw-pred-mask-prompt": [
                        wandb.Image(
                            combined_image,
                            caption="Raw | Pred | Mask | Prompt",
                        )
                    ]
                },
                commit=False,
            )

        return losses["total"]

    def forward(self, data: BatchedTomogramData) -> dict[str, Tensor]:  # type: ignore
        C, H, W = data.tomo_batch.shape[-3:]  # [H, W]
        truncate_size = 0
        # Expand channels for expected RGB input
        if C == 1:
            data.tomo_batch = data.tomo_batch.expand(-1, -1, 3, -1, -1)
            C = 3
        # Truncate if too many slices
        do_truncate = data.num_slices > MAX_SAM_DEPTH
        do_resize = self.model.image_size != H or self.model.image_size != W
        if do_truncate:
            logging.warning(
                "Truncating input tomogram from %d to %d slices for SAM2 model.",
                data.num_slices,
                MAX_SAM_DEPTH,
            )
            truncate_size = data.num_slices - MAX_SAM_DEPTH
            data.tomo_batch = data.tomo_batch[:, :MAX_SAM_DEPTH]
            data.tomo_sizes = torch.clamp(data.tomo_sizes, max=MAX_SAM_DEPTH)
            data.min_slices = min(data.min_slices, MAX_SAM_DEPTH)
        if do_resize:
            # Resize the input tomogram batch to the target size
            data.tomo_batch = F.interpolate(
                data.tomo_batch,
                size=(C, self.model.image_size, self.model.image_size),
                mode="trilinear",
                align_corners=False,
            )

        flat_tensor = data.flat_tomo_batch  # [BxDxCxHxW] -> [[BxD]xCxHxW]
        backbone_out = self.model.image_encoder(flat_tensor)
        flat_box_prompts, flat_mask_prompts = self.prompt_predictor(
            backbone_out["backbone_fpn"][0], num_batches=data.num_tomos
        )  # flat tensor form
        binary_flat_mask_prompts = (
            flat_mask_prompts > 0.9
        ).bool()  # binarize the mask prompts for SAM2 input conservatively
        preds = self.model(
            data, flat_box_prompts, binary_flat_mask_prompts
        )  # forward pass through SAM2
        masks = flat_mask_prompts.view(
            data.num_tomos, data.num_slices, *flat_mask_prompts.shape[-2:]
        )  # reshape to [B, D, H, W]

        out = {"preds": preds, "prompts": masks}

        # Pad outputs if truncated
        if do_truncate:
            pad_size = (0, 0, 0, 0, 0, truncate_size)
            for k in out:
                out[k] = F.pad(out[k], pad_size, mode="constant", value=0)
        if do_resize:
            # Upsample the output to the original size
            for k in out:
                out[k] = F.interpolate(
                    out[k], size=(H, W), mode="bilinear", align_corners=False
                )
            # Resize the input tomogram batch to the original size
            data.tomo_batch = F.interpolate(
                data.tomo_batch,
                size=(C, H, W),
                mode="trilinear",
                align_corners=False,
            )
        return out

    def load_sam_state_dict(
        self,
        state_dict: dict[str, Tensor],
        strict: bool = False,
        assign: bool = True,
    ) -> tuple:
        """Override load_state_dict to handle loading of SAM2 weights."""

        return self.model.load_state_dict(
            state_dict, strict=strict, assign=assign
        )

    def compile(self) -> None:
        """Compiles the model image encoder for training."""

        self.model.image_encoder.forward = torch.compile(
            self.model.image_encoder.forward
        )
        self.model.memory_encoder.forward = torch.compile(
            self.model.memory_encoder.forward
        )
        self.model.memory_attention.forward = torch.compile(
            self.model.memory_attention.forward
        )


class SAM2Train(SAM2Base):
    """SAMv2 model implementation."""

    def __init__(
        self,
        image_encoder: nn.Module,
        memory_attention: nn.Module,
        memory_encoder: nn.Module,
        num_init_cond_slices: tuple[int, int] = (1, 1),
        rand_init_cond_slices: tuple[bool, bool] = (True, False),
        **kwargs,
    ) -> None:
        """Initializes the SAM2 model with pre-trained blocks."""

        super().__init__(
            image_encoder, memory_attention, memory_encoder, **kwargs
        )
        self.num_init_cond_slices = num_init_cond_slices
        self.rand_init_cond_slices = rand_init_cond_slices

    def _apply_lora_to_mask_decoder(self):
        """Delay applying LoRA to the mask decoder until after loading weights."""

        decoder_factory = LoRAMaskDecoderFactory(
            lora_r=128, lora_alpha=128
        )  # Using alpha=r
        self.sam_mask_decoder = decoder_factory.apply(self.sam_mask_decoder)

    def forward(
        self, data: BatchedTomogramData, box_prompts, mask_prompts  # type: ignore
    ) -> dict[str, Any] | Tensor:
        """Forward pass for the SAMv2 model."""

        backbone_out = self.forward_image(data.flat_tomo_batch)
        mid_slice_idx = data.num_slices // 2
        backbone_out = self.prepare_prompt_inputs(
            backbone_out,
            box_prompts,
            mask_prompts,
            data,
            start_slice_idx=mid_slice_idx,
        )
        out = self.forward_tracking(backbone_out, data)
        if not isinstance(out, dict):
            out = torch.sigmoid(out)
        return out

    def prepare_prompt_inputs(
        self,
        backbone_out: dict[str, Any],
        box_prompts: dict[str, Any],
        mask_prompts: dict[str, Any],
        data: BatchedTomogramData,  # type: ignore
        start_slice_idx: int = 0,
    ) -> dict[str, Any]:
        """Prepare predicted masks."""

        backbone_out["num_slices"] = data.num_slices
        # Setup prompt parameters
        if self.training:
            num_init_cond_slices = self.num_init_cond_slices[0]
            rand_init_cond_slices = self.rand_init_cond_slices[0]
        else:
            num_init_cond_slices = self.num_init_cond_slices[1]
            rand_init_cond_slices = self.rand_init_cond_slices[1]
        assert (
            num_init_cond_slices >= 1
        ), "Number of initial conditioning slices must be at least 1."
        if rand_init_cond_slices and num_init_cond_slices > 1:
            # Randomly select number of initial conditioning slices
            num_init_cond_slices = np.random.randint(
                1, num_init_cond_slices + 1
            )

        # Select initial conditioning slices
        if num_init_cond_slices == 1:
            init_cond_slices = [start_slice_idx]
        else:
            init_cond_slices = [start_slice_idx] + np.random.choice(
                a=range(start_slice_idx + 1, data.min_slices),
                size=num_init_cond_slices - 1,
                replace=False,
            ).tolist()
        backbone_out["init_cond_slices"] = init_cond_slices
        backbone_out["slices_not_in_init_cond"] = [
            n for n in range(data.num_slices) if n not in init_cond_slices
        ]

        # Prepare mask and box inputs for slices
        backbone_out["box_inputs_per_slice"] = {}
        backbone_out["mask_inputs_per_slice"] = {}
        for n in init_cond_slices:
            idxs = data.index_to_flat_batch(n)
            backbone_out["box_inputs_per_slice"][n] = (
                box_prompts[idxs] * self.sam_prompt_encoder.input_image_size[0]
            )
            backbone_out["mask_inputs_per_slice"][n] = mask_prompts[idxs]

        return backbone_out

    def forward_tracking(
        self,
        backbone_out: dict[str, Any],
        data: BatchedTomogramData,  # type: ignore
        return_dict: bool = False,
    ) -> dict[str, Any] | Tensor:
        """Forward tracking on each slice."""

        # Prepare backbone features
        # backbone_out is [[BxD]xCxHxW]
        # vision_feats and vision_pos_embeds are [(HW), (BD), C]
        _, vision_feats, vision_pos_embeds, feat_sizes = (
            self._prepare_backbone_features(backbone_out)
        )

        # Start loop over slices
        num_slices = backbone_out["num_slices"]
        init_cond_slices = backbone_out["init_cond_slices"]
        # First process initial conditioning slices, then condition on them for memory
        processing_order = (
            init_cond_slices + backbone_out["slices_not_in_init_cond"]
        )
        # Use "frame" instead of "slice" to match with SAM2 implementation
        output_dict = {
            "cond_frame_outputs": {},
            "non_cond_frame_outputs": {},
        }
        for slice_id in processing_order:
            flat_idxs = data.index_to_flat_batch(slice_id)
            # Get image features for the current slice
            current_vision_feats = [x[:, flat_idxs] for x in vision_feats]
            current_vision_pos_embeds = [
                x[:, flat_idxs] for x in vision_pos_embeds
            ]

            current_out = self.track_step(
                frame_idx=slice_id,
                is_init_cond_frame=slice_id in init_cond_slices,
                current_vision_feats=current_vision_feats,
                current_vision_pos_embeds=current_vision_pos_embeds,
                feat_sizes=feat_sizes,
                point_inputs=backbone_out["box_inputs_per_slice"].get(
                    slice_id, None
                ),
                mask_inputs=backbone_out["mask_inputs_per_slice"].get(
                    slice_id, None
                ),
                output_dict=output_dict,
                num_frames=num_slices,
            )

            add_output_as_cond_slice = slice_id in init_cond_slices
            if add_output_as_cond_slice:
                output_dict["cond_frame_outputs"][slice_id] = current_out
            else:
                output_dict["non_cond_frame_outputs"][slice_id] = current_out
        if return_dict:
            return output_dict

        # turn 'output_dict' into a batched tensor for loss function (expects [B, D, H, W] output)
        all_slice_outputs = {}
        all_slice_outputs.update(output_dict["cond_frame_outputs"])
        all_slice_outputs.update(output_dict["non_cond_frame_outputs"])
        pred_output = []
        for _, output_dict in all_slice_outputs.items():
            # Upsample to original size (from low-res masks)
            preds = F.interpolate(
                output_dict["pred_masks"],
                scale_factor=4,
                mode="bilinear",
                align_corners=False,
            )
            pred_output.append(preds)
        total_output = torch.cat(pred_output, dim=1)

        return total_output

    def track_step(
        self,
        frame_idx: int,
        is_init_cond_frame: bool,
        current_vision_feats: Tensor | list[Tensor],
        current_vision_pos_embeds: Tensor | list[Tensor],
        feat_sizes: Tensor | list[tuple],
        point_inputs: Tensor | None,
        mask_inputs: Tensor | None,
        output_dict: dict[str, Any],
        num_frames: int,
        track_in_reverse: bool = False,
        run_mem_encoder: bool = True,
        prev_sam_mask_logits: Any | None = None,
    ) -> dict[str, Any]:
        """Process a single slice in the tomogram."""

        # Run the tracking step for the current slice
        current_out, sam_outputs, _, _ = self._track_step(
            frame_idx,
            is_init_cond_frame,
            current_vision_feats,
            current_vision_pos_embeds,
            feat_sizes,
            point_inputs,
            mask_inputs,
            output_dict,
            num_frames,
            False,
            None,
        )

        # Only save essential outputs to reduce memory usage
        (
            low_res_multimasks,
            _,
            _,
            low_res_masks,
            high_res_masks,
            obj_ptr,
            object_score_logits,
        ) = sam_outputs
        # Combine multimask outputs into one mask by taking the max
        if low_res_multimasks is not None:
            low_res_masks = torch.max(
                low_res_multimasks, dim=1, keepdim=True
            ).values
        current_out["pred_masks"] = low_res_masks
        current_out["obj_ptr"] = obj_ptr

        # Finally run the memory encoder on the predicted mask to encode
        # it into a new memory feature (that can be used in future slices)
        self._encode_memory_in_output(
            current_vision_feats,
            feat_sizes,
            None,
            True,  # run_mem_encoder
            high_res_masks,
            object_score_logits,
            current_out,
        )

        return current_out

    def _track_step(
        self,
        frame_idx,
        is_init_cond_frame,
        current_vision_feats,
        current_vision_pos_embeds,
        feat_sizes,
        point_inputs,
        mask_inputs,
        output_dict,
        num_frames,
        track_in_reverse,
        prev_sam_mask_logits,
    ):
        """Overrided _track_step to handle box and mask prompts."""

        current_out = {
            "point_inputs": None,
            "box_inputs": point_inputs,
            "mask_inputs": mask_inputs,
        }
        # High-resolution feature maps for the SAM head, reshape (HW)BC => BCHW
        if len(current_vision_feats) > 1:
            high_res_features = [
                x.permute(1, 2, 0).view(x.size(1), x.size(2), *s)
                for x, s in zip(
                    current_vision_feats[:-1], feat_sizes[:-1], strict=False
                )
            ]
        else:
            high_res_features = None
        if (
            mask_inputs is not None
            and self.use_mask_input_as_output_without_sam
        ):
            # When use_mask_input_as_output_without_sam=True, we directly output the mask input
            # (see it as a GT mask) without using a SAM prompt encoder + mask decoder.
            pix_feat = current_vision_feats[-1].permute(1, 2, 0)
            pix_feat = pix_feat.view(-1, self.hidden_dim, *feat_sizes[-1])
            sam_outputs = self._use_mask_as_output(
                pix_feat, high_res_features, mask_inputs
            )
        else:
            # fused the visual feature with previous memory features in the memory bank
            pix_feat = self._prepare_memory_conditioned_features(
                frame_idx=frame_idx,
                is_init_cond_frame=is_init_cond_frame,
                current_vision_feats=current_vision_feats[-1:],
                current_vision_pos_embeds=current_vision_pos_embeds[-1:],
                feat_sizes=feat_sizes[-1:],
                output_dict=output_dict,
                num_frames=num_frames,
                track_in_reverse=track_in_reverse,
            )
            # apply SAM-style segmentation head
            # here we might feed previously predicted low-res SAM mask logits into the SAM mask decoder,
            # e.g. in demo where such logits come from earlier interaction instead of correction sampling
            # (in this case, any `mask_inputs` shouldn't reach here as they are sent to _use_mask_as_output instead)
            if prev_sam_mask_logits is not None:
                assert point_inputs is not None and mask_inputs is None
                mask_inputs = prev_sam_mask_logits
            multimask_output = self._use_multimask(is_init_cond_frame, None)
            sam_outputs = self._forward_sam_heads(
                backbone_features=pix_feat,
                point_inputs=point_inputs,
                mask_inputs=mask_inputs,
                high_res_features=high_res_features,
                multimask_output=multimask_output,
            )

        return current_out, sam_outputs, high_res_features, pix_feat

    def _forward_sam_heads(
        self,
        backbone_features,
        point_inputs=None,
        mask_inputs=None,
        high_res_features=None,
        multimask_output=False,
    ):
        """Forward SAM prompt encoders and mask heads, overrided to use box and mask prompts."""

        B = backbone_features.size(0)
        device = backbone_features.device
        assert backbone_features.size(1) == self.sam_prompt_embed_dim
        assert backbone_features.size(2) == self.sam_image_embedding_size
        assert backbone_features.size(3) == self.sam_image_embedding_size

        # a) Handle point prompts by padding with an empty point with label -1
        sam_point_coords = torch.zeros(B, 1, 2, device=device)
        sam_point_labels = -1 * torch.ones(B, 1, device=device)

        # b) Handle mask prompts
        if mask_inputs is not None:
            # If mask_inputs is provided, downsize it into low-res mask input if needed
            # and feed it as a dense mask prompt into the SAM mask encoder
            assert len(mask_inputs.shape) == 4 and mask_inputs.shape[:2] == (
                B,
                1,
            )
            if (
                mask_inputs.shape[-2:]
                != self.sam_prompt_encoder.mask_input_size
            ):
                sam_mask_prompt = F.interpolate(
                    mask_inputs.float(),
                    size=self.sam_prompt_encoder.mask_input_size,
                    align_corners=False,
                    mode="bilinear",
                    antialias=True,  # use antialias for downsampling
                )
            else:
                sam_mask_prompt = mask_inputs
        else:
            # Otherwise, simply feed None (and SAM's prompt encoder will add
            # a learned `no_mask_embed` to indicate no mask input in this case).
            sam_mask_prompt = None

        sparse_embeddings, dense_embeddings = self.sam_prompt_encoder(
            points=(sam_point_coords, sam_point_labels),
            boxes=point_inputs,
            masks=sam_mask_prompt,
        )
        (
            low_res_multimasks,
            ious,
            sam_output_tokens,
            object_score_logits,
        ) = self.sam_mask_decoder(
            image_embeddings=backbone_features,
            image_pe=self.sam_prompt_encoder.get_dense_pe(),
            sparse_prompt_embeddings=sparse_embeddings,
            dense_prompt_embeddings=dense_embeddings,
            multimask_output=multimask_output,
            repeat_image=False,  # the image is already batched
            high_res_features=high_res_features,
        )
        if self.pred_obj_scores:
            is_obj_appearing = object_score_logits > 0

            # Mask used for spatial memories is always a *hard* choice between obj and no obj,
            # consistent with the actual mask prediction
            low_res_multimasks = torch.where(
                is_obj_appearing[:, None, None],
                low_res_multimasks,
                NO_OBJ_SCORE,
            )

        # convert masks from possibly bfloat16 (or float16) to float32
        # (older PyTorch versions before 2.1 don't support `interpolate` on bf16)
        low_res_multimasks = low_res_multimasks.float()
        high_res_multimasks = F.interpolate(
            low_res_multimasks,
            size=(self.image_size, self.image_size),
            mode="bilinear",
            align_corners=False,
        )

        sam_output_token = sam_output_tokens[:, 0]
        if multimask_output:
            # take the best mask prediction (with the highest IoU estimation)
            best_iou_inds = torch.argmax(ious, dim=-1)
            batch_inds = torch.arange(B, device=device)
            low_res_masks = low_res_multimasks[
                batch_inds, best_iou_inds
            ].unsqueeze(1)
            high_res_masks = high_res_multimasks[
                batch_inds, best_iou_inds
            ].unsqueeze(1)
            if sam_output_tokens.size(1) > 1:
                sam_output_token = sam_output_tokens[batch_inds, best_iou_inds]
        else:
            low_res_masks, high_res_masks = (
                low_res_multimasks,
                high_res_multimasks,
            )

        # Extract object pointer from the SAM output token (with occlusion handling)
        obj_ptr = self.obj_ptr_proj(sam_output_token)
        if self.pred_obj_scores:
            # Allow *soft* no obj ptr, unlike for masks
            if self.soft_no_obj_ptr:
                lambda_is_obj_appearing = object_score_logits.sigmoid()
            else:
                lambda_is_obj_appearing = is_obj_appearing.float()  # type: ignore

            if self.fixed_no_obj_ptr:
                obj_ptr = lambda_is_obj_appearing * obj_ptr
            obj_ptr = obj_ptr + (1 - lambda_is_obj_appearing) * self.no_obj_ptr

        return (
            low_res_multimasks,
            high_res_multimasks,
            ious,
            low_res_masks,
            high_res_masks,
            obj_ptr,
            object_score_logits,
        )


#### Model Creation and Loading ####


def create_sam_model_from_weights(cfg: BaseModelConfig, sam_dir: Path) -> SAM2:
    """Creates a SAM2 model from pre-trained weights specified in the config."""

    configs = _download_model_weights(sam_dir)
    assert (
        cfg.name in configs
    ), f"Model {cfg.name} was not found in available SAMv2 models. Available models are {configs.keys()}."

    file_paths = configs[cfg.name]

    # Merge configs together
    model_cfg_path = file_paths["config"]
    model_cfg = OmegaConf.load(model_cfg_path)["model"]  # type: ignore
    model_cfg._target_ = (
        "cryovit.models.sam2.SAM2Train"  # Use cryovit SAM2 as target
    )
    model_cfg.image_size = (
        512  # Set image size to 512 (crop size for training)
    )
    model_cfg.use_mask_input_as_output_without_sam = (
        False  # use sam memory and mask decoder
    )
    model_cfg._partial_ = True

    model = instantiate(
        cfg, sam_model=model_cfg, custom_kwargs=cfg.custom_kwargs
    )
    sd = torch.load(
        file_paths["weights"], map_location="cpu", weights_only=True
    )["model"]
    missing_keys, unexpected_keys = model.load_sam_state_dict(sd)
    if missing_keys:
        logging.error(missing_keys)
        raise RuntimeError()
    if unexpected_keys:
        logging.error(unexpected_keys)
        raise RuntimeError()
    model.model._apply_lora_to_mask_decoder()  # Apply LoRA after loading weights
    model.configure_optimizers()  # Configure optimizers after setting requires_grad

    return model


def _download_model_weights(sam_dir: Path) -> dict[str, dict[str, Path]]:
    """Downloads the SAMv2 and Medical-SAMv2 model weights if they do not exist using huggingface_hub."""

    # Download base SAMv2 model
    sam2_repo, sam2_config = sam2_model
    if not (
        (sam_dir / sam2_config["weights"]).exists()
        and (sam_dir / sam2_config["config"]).exists()
    ):
        from huggingface_hub import snapshot_download

        snapshot_download(
            repo_id=sam2_repo, repo_type="model", local_dir=sam_dir
        )
    sam2_config = {k: sam_dir / v for k, v in sam2_config.items()}

    # Download Medical-SAMv2
    medsam_repo, medsam_config = medical_sam2_model
    if not (
        (sam_dir / medsam_config["weights"]).exists()
        and (sam_dir / medsam_config["config"]).exists()
    ):
        from huggingface_hub import snapshot_download

        snapshot_download(
            repo_id=medsam_repo, repo_type="model", local_dir=sam_dir
        )
    medsam_config = {k: sam_dir / v for k, v in medsam_config.items()}

    return {"SAM2": sam2_config, "MedSAM": medsam_config}



"""Functions for plotting and visualizing results from CryoViT experiments."""

from cryovit.visualization.dino_pca import process_samples
from cryovit.visualization.fractional_sample import (
    process_fractional_experiment,
)
from cryovit.visualization.multi_label_sample import (
    process_multi_label_experiment,
)
from cryovit.visualization.multi_sample import process_multi_experiment
from cryovit.visualization.segmentations import process_experiment
from cryovit.visualization.single_sample import process_single_experiment
from cryovit.visualization.sparse_sample import process_sparse_experiment

__all__ = [
    "process_samples",
    "process_experiment",
    "process_single_experiment",
    "process_multi_experiment",
    "process_multi_label_experiment",
    "process_fractional_experiment",
    "process_sparse_experiment",
]



"""Base Lightning Module class for 3D Tomogram Segmentation."""

from abc import ABC, abstractmethod
from collections.abc import Callable
from copy import deepcopy
from typing import Literal

import torch
from pytorch_lightning import LightningModule
from pytorch_lightning.utilities import grad_norm
from torch import Tensor, nn
from torch.optim import Optimizer

from cryovit.types import BatchedModelResult, BatchedTomogramData


class BaseModel(LightningModule, ABC):
    """Base model with configurable loss functions and metrics."""

    def __init__(
        self,
        input_key: str,
        lr: float,
        weight_decay: float,
        losses: dict[str, Callable],
        metrics: dict[str, Callable],
        name: str = "BaseModel",
        custom_kwargs: dict | None = None,
        **kwargs,
    ) -> None:
        """Initializes the BaseModel with specified learning rate, weight decay, loss functions, and metrics.

        Args:
            input_key (str): Key to access input data in the batch.
            lr (float): Learning rate for the optimizer.
            weight_decay (float): Weight decay factor for AdamW optimizer.
            losses (dict[str, Callable]): Dictionary of loss functions for training, validation, and testing.
            metrics (dict[str, Callable]): Dictionary of metric functions for training, validation, and testing.
            name (str): Name of the model.
            custom_kwargs (Optional[dict[str, Any]]): Additional custom keyword arguments to set as attributes.
        """

        super().__init__()
        self.name = name
        self.input_key = input_key
        self.lr = lr
        self.weight_decay = weight_decay

        if custom_kwargs is not None:
            for key, value in custom_kwargs.items():
                setattr(self, key, value)

        self.configure_losses(losses)
        self.configure_metrics(metrics)

        self.save_hyperparameters()

    def configure_optimizers(self) -> Optimizer:
        return torch.optim.AdamW(
            self.parameters(),
            lr=self.lr,
            weight_decay=self.weight_decay,
        )

    def configure_losses(self, losses: dict[str, Callable]) -> None:
        self.loss_fns = losses

    def configure_metrics(self, metrics: dict[str, Callable]) -> None:
        self.metric_fns = nn.ModuleDict(
            {
                "TRAIN": nn.ModuleDict({m: deepcopy(m_fn) for m, m_fn in metrics.items()}),  # type: ignore
                "VAL": nn.ModuleDict({m: deepcopy(m_fn) for m, m_fn in metrics.items()}),  # type: ignore
                "TEST": nn.ModuleDict({m: deepcopy(m_fn) for m, m_fn in metrics.items()}),  # type: ignore
            }
        )

    def on_before_optimizer_step(self, optimizer: Optimizer) -> None:
        """Logs gradient norms just before the optimizer updates weights."""

        norms = grad_norm(self, norm_type=2)
        self.log_dict(norms, on_step=True)

    def _masked_predict(
        self, batch: BatchedTomogramData, use_mito_mask: bool = False  # type: ignore
    ) -> dict[str, Tensor]:
        """Performs prediction while applying a mask to the inputs and labels based on the label value."""

        y_true = batch.labels  # (B, D, H, W)

        y_pred_full = self(batch)  # (B, D, H, W)
        mask = (y_true > -1.0).detach()
        if use_mito_mask:
            assert (
                batch.aux_data is not None and "labels/mito" in batch.aux_data
            ), "Batch aux_data must contain 'labels/mito' key for mito masking."
            # assumes eval code with batch size of 1
            mito_mask = torch.from_numpy(batch.aux_data["labels/mito"][0]) > 0
            mito_mask = mito_mask.to(dtype=mask.dtype, device=mask.device)
            mask = mask & mito_mask  # Combine masks

        y_pred = torch.masked_select(y_pred_full, mask).view(-1, 1)
        y_true = torch.masked_select(y_true, mask).view(-1, 1)

        return {"preds": y_pred, "labels": y_true, "preds_full": y_pred_full}

    def compute_losses(
        self, y_pred: Tensor, y_true: Tensor
    ) -> dict[str, Tensor]:
        losses = {k: v(y_pred, y_true) for k, v in self.loss_fns.items()}
        losses["total"] = sum(losses.values())
        return losses

    def log_stats(
        self,
        losses: dict[str, Tensor],
        prefix: Literal["train", "val", "test"],
        batch_size: int,
    ) -> None:
        """Logs computed loss and metric statistics for each training or validation step."""

        # Log losses
        loss_log_dict = {f"{prefix}/loss/{k}": v for k, v in losses.items()}
        on_step = prefix == "train"
        self.log_dict(
            loss_log_dict,
            prog_bar=True,
            on_epoch=not on_step,
            on_step=on_step,
            batch_size=batch_size,
        )

        # Log metrics
        metric_log_dict = {}
        for m, m_fn in self.metric_fns[prefix.upper()].items():  # type: ignore
            metric_log_dict[f"{prefix}/metric/{m}"] = m_fn
        self.log_dict(
            metric_log_dict,
            prog_bar=True,
            on_epoch=True,
            on_step=False,
            batch_size=batch_size,
        )

    def _do_step(self, batch: BatchedTomogramData, batch_idx: int, prefix: Literal["train", "val", "test"]) -> Tensor:  # type: ignore
        """Processes a single batch of data, computes the loss and updates metrics."""

        out = self._masked_predict(batch)
        y_pred, y_true = out["preds"], out["labels"]

        losses = self.compute_losses(y_pred, y_true)

        for _, m_fn in self.metric_fns[prefix.upper()].items():  # type: ignore
            m_fn(y_pred, y_true)

        self.log_stats(losses, prefix, batch.num_tomos)
        return losses["total"]

    def training_step(self, batch: BatchedTomogramData, batch_idx: int) -> Tensor:  # type: ignore
        """Processes one batch during training, returning the total loss."""

        return self._do_step(batch, batch_idx, "train")  # type: ignore

    def validation_step(self, batch: BatchedTomogramData, batch_idx: int) -> Tensor:  # type: ignore
        """Processes one batch during validation, returning the total loss."""

        return self._do_step(batch, batch_idx, "val")  # type: ignore

    def test_step(self, batch: BatchedTomogramData, batch_idx: int) -> BatchedModelResult:  # type: ignore
        """Processes one batch during testing, captures predictions, and computes losses and metrics.

        Args:
            batch (BatchedTomogramData): The batch of data being processed.
            batch_idx (int): Index of the batch.

        Returns:
            BatchedModelResult: Contains test results and metrics, as well as file metadata for this batch.
        """

        assert (
            batch.aux_data is not None and "data" in batch.aux_data
        ), "Batch aux_data must contain 'data' key for testing."
        input_data = batch.aux_data["data"]

        # whether to use mito labels to mask granule predictions in evaluation
        use_mito_mask = (
            "labels/mito" in batch.aux_data and batch.aux_data["labels/mito"]
            if batch.aux_data is not None
            else False
        )
        out_dict = self._masked_predict(batch, use_mito_mask=use_mito_mask)
        y_pred, y_true, y_pred_full = (
            out_dict["preds"],
            out_dict["labels"],
            out_dict["preds_full"],
        )
        aux_data = {}
        for key in out_dict:
            if key not in ["preds", "labels", "preds_full"] and key.endswith(
                "full"
            ):
                aux_data_full = out_dict[key]
                aux_data[key] = [
                    t_aux_data.cpu().numpy() for t_aux_data in aux_data_full
                ]

        samples, tomo_names = batch.metadata.identifiers
        split_id = batch.metadata.split_id
        labels = [t_labels.cpu().numpy() for t_labels in batch.labels]
        preds = [t_preds.float().cpu().numpy() for t_preds in y_pred_full]
        if split_id is not None:
            split_id = [s_id.item() for s_id in split_id]

        losses = {
            k: v.item() for k, v in self.compute_losses(y_pred, y_true).items()
        }
        metrics = {}
        for m, m_fn in self.metric_fns["TEST"].items():  # type: ignore
            score = m_fn(y_pred, y_true)
            metrics[m] = score.item()
            m_fn.reset()

        return BatchedModelResult(
            num_tomos=batch.num_tomos,
            samples=samples,
            tomo_names=tomo_names,
            split_id=split_id,
            data=input_data,
            label=labels,
            preds=preds,  # type: ignore
            losses=losses,
            metrics=metrics,
            aux_data=aux_data if aux_data else None,
        )

    def predict_step(self, batch: BatchedTomogramData, batch_idx: int) -> BatchedModelResult:  # type: ignore
        """Processes one batch during prediction, capturing model outputs along with file metadata.

        Args:
            batch (BatchedTomogramData): The batch of data being processed.
            batch_idx (int): Index of the batch.

        Returns:
            BatchedModelResult: Contains test results and metrics, as well as file metadata for this batch.
        """

        assert (
            batch.aux_data is not None and "data" in batch.aux_data
        ), "Batch aux_data must contain 'data' key for prediction."
        input_data = batch.aux_data["data"]
        preds = self(batch)
        labels = [t_labels.cpu().numpy() for t_labels in batch.labels]
        samples, tomo_names = batch.metadata.identifiers

        return BatchedModelResult(
            num_tomos=batch.num_tomos,
            samples=samples,
            tomo_names=tomo_names,
            split_id=None,
            data=input_data,
            label=labels,
            preds=[t_preds.float().cpu().numpy() for t_preds in preds],
            losses={},
            metrics={},
            aux_data=None,
        )

    @abstractmethod
    def forward(self):
        """Should be implemented in subclass."""

        raise NotImplementedError(
            "The forward method must be implemented by subclass."
        )



"""Loss functions for training CryoVIT segmentation models."""

import torch
from torch import Tensor, nn
from torchvision.ops import sigmoid_focal_loss


class DiceLoss(nn.Module):
    """Dice Loss for imbalanced foreground / background segmentation tasks."""

    def __init__(self) -> None:
        """Initializes the DiceLoss instance."""

        super().__init__()
        self.name = "DiceLoss"

    def forward(self, y_pred: Tensor, y_true: Tensor) -> Tensor:
        """Computes the Dice loss between predictions and true values.

        Args:
            y_pred (Tensor): Predicted probabilities, expected to be logits from the model.
            y_true (Tensor): Ground truth labels.

        Returns:
            Tensor: Computed Dice loss value.
        """

        intersection = torch.sum(y_true * y_pred)
        denom = torch.sum(y_true) + torch.sum(y_pred)
        dice_loss = 1 - (2 * intersection) / (denom + 1e-3)

        return dice_loss


class FocalLoss(nn.Module):
    """Focal loss to address class imbalance by focusing more on hard-to-classify instances."""

    def __init__(self, gamma=2, **kwargs) -> None:
        """Initializes the FocalLoss instance.

        Args:
            gamma (float, optional): Focusing parameter. Default is 2.
        """

        super().__init__()
        self.gamma = gamma
        self.name = "FocalLoss"

    def forward(self, y_pred: Tensor, y_true: Tensor) -> Tensor:
        """Computes the Focal loss between predictions and true values.

        Args:
            y_pred (Tensor): Predicted probabilities, expected to be logits from the model.
            y_true (Tensor): Ground truth labels.

        Returns:
            Tensor: Computed Focal loss value.
        """

        weight = (y_true.numel() - y_true.sum()) / y_true.numel()
        return sigmoid_focal_loss(
            y_pred,
            y_true,
            alpha=weight.item(),
            gamma=self.gamma,
            reduction="mean",
        )



from collections.abc import Callable
from pathlib import Path

import pandas as pd
from scipy.stats import ttest_rel, wilcoxon


def merge_experiments(
    exp_dir: Path,
    exp_names: dict[str, list[str]],
    keys: list[str] | None = None,
) -> pd.DataFrame:
    """Merge multiple experiment results into a single DataFrame, adding additional labels for each experiment.

    Args:
        exp_dir (Path): The directory containing experiment results (.csvs).
        exp_names (dict[str, list[str]]): A dictionary mapping experiment names to labels.
        keys (list[str]): The column names to assign to the experiment labels in the merged DataFrame. If None, defaults to ["model"].

    Raises:
        ValueError: If the specified experiment directory does not exist.

    Returns:
        pd.DataFrame: A DataFrame containing merged experiment data.
    """

    if not exp_dir.exists():
        raise ValueError(f"The directory {exp_dir} does not exist")

    if keys is None:
        keys = ["model"]
    results = []

    for exp_name, value in exp_names.items():
        result_csv = exp_dir / f"{exp_name}.csv"
        df = pd.read_csv(result_csv)
        for key, val in zip(keys, value, strict=True):
            df[key] = val
        results.append(df)

    return pd.concat(results, axis=0, ignore_index=True)


def significance_test(
    df,
    model_A: str,
    model_B: str,
    key: str = "model",
    test_fn: str = "wilcoxon",
) -> float:
    """Perform a significance test with a specific function between two models on a grouped DataFrame.

    Args:
        group (pd.DataFrame): Grouped DataFrame to perform the test on.
        model_A (str): The name of the first model.
        model_B (str): The name of the second model.
        key (str): Column name to get model names.
        test_fn (str): The statistical test function to use (e.g., wilcoxon, ttest_rel).

    Raises:
        ValueError: If an unknown test function is provided.

    Returns:
        float: The p-value from the statistical test.
    """

    score_A = df[df[key] == model_A].sort_values("tomo_name").dice_metric
    score_B = df[df[key] == model_B].sort_values("tomo_name").dice_metric
    assert len(score_A) == len(
        score_B
    ), "The two models must have the same number of samples for comparison."

    if test_fn == "wilcoxon":
        _, pvalue = wilcoxon(
            score_A, score_B, method="exact", alternative="two-sided"
        )
    elif test_fn == "ttest_rel":
        _, pvalue = ttest_rel(score_A, score_B, alternative="two-sided")
    else:
        raise ValueError(f"Unknown test function: {test_fn}")

    return pvalue  # type: ignore


def compute_stats(
    df: pd.DataFrame, group_keys: list[str], file_name: str, test_fn: Callable
) -> pd.Series:
    """Compute statistical summaries for the DataFrame and save them to a file.

    Args:
        df (pd.DataFrame): The DataFrame to compute statistics on.
        group_keys (list[str]): The column names used to group data for statistics. The first element is used for the p-value calculation.
        file_name (str): The file path to save the statistics.
        test_fn (Callable): A function to compute p-values between groups.

    Returns:
        pd.Series: A Series containing p-values for statistical tests.
    """

    grouped = df.groupby(group_keys, sort=False)["dice_metric"].agg(
        mean="mean",
        std="std",
        median="median",
        Q1=lambda x: x.quantile(0.25),
        Q3=lambda x: x.quantile(0.75),
    )

    transforms = {
        "Median Dice Score": lambda row: f"{row['median']:.2f}",
        "Mean Dice Score ± Std": lambda row: f"{row['mean']:.2f} ± {row['std']:.2f}",
        "Dice Score Quartiles (Q1 - Q3)": lambda row: f"{row['Q1']:.2f} - {row['Q3']:.2f}",
    }

    values = {
        col: grouped.apply(func, axis=1) for col, func in transforms.items()
    }
    stats_df = pd.DataFrame.from_dict(values).unstack(level=-1)

    pvalues = df.groupby(group_keys[0]).apply(test_fn, include_groups=False)
    pvalues_formatted = pvalues.apply(lambda x: f"{x:.2e}")
    stats_df["p-value"] = pvalues_formatted[stats_df.index]

    if (
        group_keys[0] != "split_id"
    ):  # Don't count n for splits grouping (i.e., fractional sample)
        sample_counts = df[group_keys[0]].value_counts(ascending=True)
        stats_df = stats_df.loc[sample_counts.index]
    stats_df.reset_index(names=group_keys[0]).to_csv(file_name, index=False)  # type: ignore

    return pvalues



"""Implementations of deep-learning automated segmentation models for CryoViT."""

from cryovit.models.base_model import BaseModel
from cryovit.models.cryovit import CryoVIT
from cryovit.models.sam2 import SAM2, create_sam_model_from_weights
from cryovit.models.unet3d import UNet3D

__all__ = [
    "BaseModel",
    "CryoVIT",
    "UNet3D",
    "SAM2",
    "create_sam_model_from_weights",
]



"""Extract visualizations of DINO features with PCA."""

import logging
from pathlib import Path

import h5py
import numpy as np
import torch
import torch.nn.functional as F
from numpy.typing import NDArray
from rich.progress import track

from cryovit.config import tomogram_exts


def _calculate_pca(features: NDArray[np.float16]) -> NDArray[np.float32]:
    from sklearn.decomposition import PCA
    from umap import UMAP

    float_features = features.astype(np.float32)  # type: ignore # PCA expects float32
    x = float_features.transpose((1, 2, 3, 0))
    x = x.reshape((-1, x.shape[-1]))  # N, C
    # Reduce dimensionality to 3 colors
    pca = PCA(n_components=min(1024, x.shape[0]))
    x = pca.fit_transform(x)
    umap = UMAP(n_components=3, verbose=False, n_jobs=16)
    umap.fit(x)
    # Upscale features
    torch_features = F.interpolate(
        torch.from_numpy(features), scale_factor=2, mode="bicubic"
    )
    D, W, H = torch_features.shape[1:]  # D, W, H
    torch_features = torch_features.permute(1, 2, 3, 0).contiguous()
    np_features = torch_features.view(-1, torch_features.shape[-1]).numpy()
    np_features = pca.transform(np_features)
    np_features = umap.transform(np_features)
    return np_features.reshape(D, W, H, 3)  # type: ignore


def _color_features(
    features: NDArray[np.float32], alpha: float = 0.0
) -> NDArray[np.uint8]:
    from matplotlib.colors import hsv_to_rgb, rgb_to_hsv

    # Normalize
    features = features - features.min(axis=(0, 1, 2))
    features = features / features.max(axis=(0, 1, 2))

    # Normalize colors
    hsv = rgb_to_hsv(features)
    hsv[..., 1] = 0.9
    hsv[..., 2] = 0.75
    hsv[..., 0] = (alpha + hsv[..., 0]) % 1.0  # alpha = 0
    rgb = hsv_to_rgb(hsv)
    rgb = (255 * rgb).astype(np.uint8)

    # Upscale to full image size
    rgb = np.repeat(rgb, 8, axis=1)
    rgb = np.repeat(rgb, 8, axis=2)
    return rgb  # type: ignore


def export_pca(
    data: NDArray[np.float32],
    features: NDArray[np.float16],
    tomo_name: str,
    result_dir: Path,
    frame_id: int | None = None,
) -> None:
    """Extract PCA colormap from features and save to a specified directory."""

    from PIL import Image

    # Save as Images
    image_dir = result_dir / tomo_name
    image_dir.mkdir(parents=True, exist_ok=True)

    if frame_id is None:
        idxs = list(np.arange(0, data.shape[0], step=10, dtype=int))
    else:
        idxs = [frame_id]
    np_features = features[:, idxs]
    np_features = _calculate_pca(np_features)
    np_features = _color_features(np_features)
    for i, idx in enumerate(idxs):  # type: ignore
        img_path = image_dir / f"{idx}.png"

        data = data - data.min()
        data = data / data.max()
        int_data = (data * 255.0).astype(np.uint8)
        f_img = Image.fromarray(np_features[i][::-1])
        d_img = Image.fromarray(int_data[idx][::-1])

        img = Image.new(
            "RGB", (2 * f_img.size[0], f_img.size[1])
        )  # concat images
        img.paste(d_img)
        img.paste(f_img, box=(d_img.size[0], 0))
        logging.debug("Saving PCA visualization to %s", img_path)
        img.save(img_path)


def process_samples(
    exp_dir: Path, result_dir: Path, sample: str | None = None
):
    """Process all samples in an experiment directory and save PCA visualizations."""

    result_dir.mkdir(parents=True, exist_ok=True)
    samples = (
        [s.name for s in exp_dir.iterdir() if s.is_dir()]
        if sample is None
        else [sample]
    )
    logging.info(
        "Found %d samples in experiment directory %s: %s",
        len(samples),
        exp_dir,
        samples,
    )

    for sample in samples:
        tomo_dir = exp_dir / sample
        tomo_names = [
            f.name for f in tomo_dir.glob("*") if f.suffix in tomogram_exts
        ]
        for tomo_name in track(
            tomo_names,
            description=f"[green]Processing {sample}",
            total=len(tomo_names),
        ):
            with h5py.File(tomo_dir / tomo_name) as fh:
                data: NDArray[np.float32] = fh["data"][()]  # type: ignore
                if data.dtype == np.uint8:
                    data = data.astype(np.float32) / 255.0
                features: NDArray[np.float16] = fh["dino_features"][()].astype(np.float32)  # type: ignore
                export_pca(data, features, tomo_name[:-4], result_dir / sample)



from pathlib import Path
from typing import Annotated

from typer import Argument, Option

from cryovit.types import ModelType

from .cli import cli


@cli.command(name="train", no_args_is_help=True)
def train(
    train_data: Annotated[
        str,
        Argument(
            help="Path to the folder or .txt file containing the training tomograms.",
        ),
    ],
    train_labels: Annotated[
        str,
        Argument(
            help="Path to the folder or .txt file containing the training labels.",
        ),
    ],
    labels: Annotated[
        list[str],
        Option(help="List of available label names in ascending-value order."),
    ],
    label_key: Annotated[
        str,
        Argument(
            help="Label key to train on. Must be one of the provided labels.",
        ),
    ],
    validation_data: Annotated[
        str | None,
        Option(
            "--validation-data",
            "--val-data",
            help="Path to the folder or .txt file containing the validation tomograms.",
            show_default=False,
        ),
    ] = None,
    validation_labels: Annotated[
        str | None,
        Option(
            "--validation-labels",
            "--val-labels",
            help="Path to the folder or .txt file containing the validation labels.",
            show_default=False,
        ),
    ] = None,
    name: Annotated[
        str | None,
        Option(
            help="Name to identify the model. If not provided, a random name will be generated.",
            show_default=False,
            rich_help_panel="Model Customization",
        ),
    ] = None,
    model: Annotated[
        ModelType,
        Option(
            help="The type of model to train.",
            rich_help_panel="Model Customization",
        ),
    ] = ModelType.CRYOVIT,
    result_folder: Annotated[
        str | None,
        Option(
            help="Path to the folder where the trained model will be saved.",
            show_default="the current working directory",
            rich_help_panel="Model Customization",
        ),
    ] = None,
    ckpt: Annotated[
        str | None,
        Option(
            help="Path to a pre-trained .model file, or .ckpt/.pt weights to fine-tune a model from.",
            rich_help_panel="Model Customization",
        ),
    ] = None,
    num_epochs: Annotated[
        int,
        Option(
            min=1,
            help="Number of training epochs.",
            rich_help_panel="Training Customization",
        ),
    ] = 50,
    log_training: Annotated[
        bool,
        Option(
            "--log-training",
            "-l",
            help="Additionally log training metrics to TensorBoard?",
            rich_help_panel="Training Customization",
        ),
    ] = False,
):
    """Train a segmentation model on given training data and labels.

    Example
    -------
    cryovit train <path-to-training-data> <path-to-training-labels> --labels <first-label> <second-label> ... --label-key <label-to-train-on>
    """
    from cryovit._logging_config import setup_logging
    from cryovit.run.train_model import run_training
    from cryovit.utils import (
        id_generator,
        load_files_from_path,
    )

    setup_logging("INFO")

    ## Convert Arguments
    train_path = Path(train_data)
    label_path = Path(train_labels)
    ckpt_path = Path(ckpt) if ckpt is not None else None
    val_path = Path(validation_data) if validation_data else None
    val_label_path = Path(validation_labels) if validation_labels else None
    result_path = Path(result_folder) if result_folder else Path.cwd()
    model_name = name or model.value + "_" + id_generator()

    ## Sanity Checking
    assert train_path.exists(), "Training data path does not exist."
    assert label_path.exists(), "Training labels path does not exist."
    if val_path is not None:
        assert val_path.exists(), "Validation data path does not exist."
        assert (
            val_label_path is not None and val_label_path.exists()
        ), "Validation data provided but validation labels path does not exist."
    if ckpt_path is not None:
        assert ckpt_path.exists(), "Checkpoint path does not exist."

    train_files = load_files_from_path(train_path)
    train_label_files = load_files_from_path(label_path)
    val_files = (
        load_files_from_path(val_path) if val_path is not None else None
    )
    val_label_files = (
        load_files_from_path(val_label_path)
        if val_label_path is not None
        else None
    )
    result_path.mkdir(parents=True, exist_ok=True)

    run_training(
        train_files,
        train_label_files,
        labels,
        model,
        model_name,
        label_key,
        result_path,
        val_files,
        val_label_files,
        num_epochs=num_epochs,
        log_training=log_training,
        ckpt_path=ckpt_path,
    )



import logging
from pathlib import Path
from typing import Any

import h5py
import numpy as np
import pandas as pd
from pytorch_lightning import Callback, LightningModule, Trainer
from pytorch_lightning.callbacks.prediction_writer import BasePredictionWriter
from pytorch_lightning.utilities.types import STEP_OUTPUT

from cryovit.types import BatchedModelResult


class TestPredictionWriter(Callback):
    """Callback to write predictions to disk during model evaluation."""

    def __init__(self, results_dir: Path, label_key: str, **kwargs) -> None:
        """Creates a callback to save predictions on the test data.

        Args:
            results_dir (Path): directory in which the predictions should be saved.
        """

        self.results_dir = (
            results_dir if isinstance(results_dir, Path) else Path(results_dir)
        )
        self.label_key = label_key

    def on_test_batch_end(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        outputs: STEP_OUTPUT,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        """Handles the end of a test batch to save outputs."""

        assert isinstance(outputs, BatchedModelResult)
        for n in range(outputs.num_tomos):
            output_file = (
                self.results_dir / outputs.samples[n] / outputs.tomo_names[n]
            )
            output_file.parent.mkdir(parents=True, exist_ok=True)

            # Binary classify predictions into segmentations and convert to correct formats
            data = outputs.data[n]
            labels = outputs.label[n]
            preds = outputs.preds[n]

            with h5py.File(output_file, "w") as fh:
                fh.create_dataset(
                    "data", data=data, shape=data.shape, dtype=data.dtype
                )
                fh.create_dataset(self.label_key, data=labels, shape=labels.shape, dtype=labels.dtype, compression="gzip")  # type: ignore
                fh.create_dataset(f"{self.label_key}_preds", data=preds, shape=preds.shape, dtype=preds.dtype, compression="gzip")  # type: ignore

                if outputs.aux_data is not None:
                    fh.create_group("aux_data")
                    for key in outputs.aux_data:
                        fh["aux_data"].create_dataset(key, data=outputs.aux_data[key], compression="gzip")  # type: ignore


class PredictionWriter(BasePredictionWriter):
    """Callback to write predictions to disk during model prediction."""

    def __init__(
        self, results_dir: Path, label_key: str, threshold: float, **kwargs
    ) -> None:
        """Creates a callback to save predictions on the test data.

        Args:
            results_dir (Path): directory in which the predictions should be saved.
        """

        super().__init__(write_interval="batch", **kwargs)
        self.results_dir = (
            results_dir if isinstance(results_dir, Path) else Path(results_dir)
        )
        self.label_key = label_key
        self.threshold = threshold
        self.result_paths = []

    def write_on_batch_end(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        prediction: BatchedModelResult,
        batch_indices,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        """Handles the end of a prediction batch to save outputs."""

        for n in range(prediction.num_tomos):
            result_path = self.results_dir / prediction.tomo_names[n]
            result_path = result_path.with_suffix(".hdf")
            result_path.parent.mkdir(parents=True, exist_ok=True)
            # Binary classify predictions into segmentations and convert to correct formats
            data = prediction.data[n].astype(np.float32)
            preds = prediction.preds[n]
            segs = (preds >= self.threshold).astype(
                np.uint8
            )  # Binary segmentation
            # Save data and pred to HDF5
            with h5py.File(result_path, "w") as fh:
                fh.create_dataset("data", data=data, compression="gzip")
                fh.create_dataset(
                    f"{self.label_key}_preds", data=segs, compression="gzip"
                )
            self.result_paths.append(result_path)


class CsvWriter(Callback):
    """Callback to save model performance metrics to a .csv."""

    def __init__(self, csv_result_path: Path, **kwargs) -> None:
        """Creates a callback to save performance metrics on the test data.

        Args:
            csv_result_path (Path): .csv file in which metrics should be saved.
        """

        self.csv_result_path = (
            csv_result_path
            if isinstance(csv_result_path, Path)
            else Path(csv_result_path)
        )
        self.csv_result_path.parent.mkdir(parents=True, exist_ok=True)

    def _create_metric_df(
        self,
        sample: str,
        tomo_name: str,
        split_id: int | None = None,
        **metrics,
    ) -> pd.DataFrame:
        result_dict: dict[str, Any] = {
            "sample": sample,
            "tomo_name": tomo_name,
        }
        for name, value in metrics.items():
            result_dict[name] = [value]
        if split_id is not None:
            result_dict["split_id"] = [split_id]

        return pd.DataFrame(result_dict)

    def on_test_batch_end(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        outputs: STEP_OUTPUT,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        """Handles the end of a test batch to save metrics."""

        assert isinstance(outputs, BatchedModelResult)
        assert (
            outputs.num_tomos == 1
        ), "TestPredictionWriter only supports single-tomogram batches."
        sample, tomo_name, split_id = (
            outputs.samples[0],
            outputs.tomo_names[0],
            outputs.split_id[0] if outputs.split_id is not None else None,
        )

        # Create results .csv if it doesn't exist
        metric_names = list(outputs.metrics)
        column_names = ["sample", "tomo_name"] + metric_names
        if split_id is not None:
            column_names += ["split_id"]
        if not self.csv_result_path.exists():
            results_df = pd.DataFrame(columns=column_names)
        else:
            results_df = pd.read_csv(self.csv_result_path)

        # Warn if row already exists and remove (i.e., replace)
        matching_rows = (results_df["tomo_name"] == tomo_name) & (
            results_df["sample"] == sample
        )
        if split_id is not None and "split_id" in results_df.columns:
            matching_rows = matching_rows & (
                results_df["split_id"] == split_id
            )
        if matching_rows.any():
            logging.warning(
                "Data with sample %s, name %s, and split %s already has an entry. Replacing %d rows...",
                sample,
                tomo_name,
                split_id,
                matching_rows.sum(),
            )
            results_df = results_df[~matching_rows]
        # Add metrics to df
        metrics_df = self._create_metric_df(
            sample, tomo_name, split_id, **outputs.metrics
        )
        if results_df.empty:
            results_df = metrics_df
        else:
            results_df = pd.concat([results_df, metrics_df], ignore_index=True)
        results_df.to_csv(self.csv_result_path, mode="w", index=False)



"""Implementation of the multi sample data module."""

import pandas as pd

from cryovit.datamodules.base_datamodule import BaseDataModule


class MultiSampleDataModule(BaseDataModule):
    """Data module for CryoVIT experiments involving multiple samples."""

    def __init__(
        self,
        sample: list[str],
        split_id: int | None,
        split_key: str | None,
        test_sample: list[str] | None = None,
        **kwargs,
    ) -> None:
        """Train on a fraction of tomograms and leave out one sample for evaluation.

        Args:
            sample (list[str]): list of samples used for training.
            split_id (Optional[int]): An optional split ID for validation.
            split_key (str): The key used to select splits using split_id.
            test_sample (Optional[list[str]]): list of samples used for testing. If None, test on the validation set.
        """

        super().__init__(**kwargs)
        # Validity checks
        assert isinstance(
            sample, list
        ), f"Multi sample 'sample' should be a list. Got {sample} instead."
        assert test_sample is None or isinstance(
            test_sample, list
        ), f"Multi sample 'test_sample' should be None or a list. Got {test_sample} instead."

        self.sample = sample
        self.split_id = split_id
        self.split_key = split_key
        self.test_sample = test_sample

    def train_df(self) -> pd.DataFrame:
        """Train tomograms: exclude those with the specified split_id.

        Returns:
            pd.DataFrame: A dataframe specifying the train tomograms.
        """

        assert self.record_df is not None
        if self.split_id is not None:
            df = self.record_df[
                (self.record_df[self.split_key] != self.split_id)
                & (self.record_df["sample"].isin(self.sample))
            ]
        else:
            df = self.record_df[self.record_df["sample"].isin(self.sample)][
                ["sample", "tomo_name"]
            ]

        return df

    def val_df(self) -> pd.DataFrame:
        """Validation tomograms: optionally validate on tomograms with the specified split_id.

        Returns:
            pd.DataFrame: A dataframe specifying the validation tomograms.
        """

        assert self.record_df is not None
        if self.split_id is None:  # validate on train set
            return self.train_df()

        return self.record_df[
            (self.record_df[self.split_key] == self.split_id)
            & (self.record_df["sample"].isin(self.sample))
        ]

    def test_df(self) -> pd.DataFrame:
        """Test tomograms: test on tomograms from the test samples.

        Returns:
            pd.DataFrame: A dataframe specifying the test tomograms.
        """

        assert self.record_df is not None
        if self.test_sample is None:
            return self.val_df()

        return self.record_df[
            (self.record_df["sample"].isin(self.test_sample))
        ][["sample", "tomo_name"]]

    def predict_df(self) -> pd.DataFrame:
        """Predict tomograms: predict on the specified samples.

        Returns:
            pd.DataFrame: A dataframe specifying the predict tomograms.
        """

        assert self.record_df is not None
        return self.record_df[self.record_df["sample"].isin(self.sample)][
            ["sample", "tomo_name"]
        ]
